/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if !defined(KCanServer_H)

  #define KCanServer_H

  #include "NCanUtils.h"
  #include "KCanDriver.h"
  #include "KThread.h"

  class KCanServer : public KThread 
  {
    private:
      unsigned FrameCount;
      bool IsInitalized;
      class KCriticalSection * Semaphore;
      int RingBufferSize;
      volatile int RingHead;
      volatile int RingTail;
      KCanFrame * RingBuffer;
      KCanDriver * CanDriver;
      KCanDriver * CanSendDriver;
    
      void IncRingPtr(volatile int & RingPtr);
    public:
      enum { dt_can, dt_can232, dt_cs } driver_type;
    
      KCanServer();
      ~KCanServer();
  
      bool Init(const char * CanDev);
      void Halt();
      void PutFrame(const KCanFrame & Frame);
      bool GetFrame(KCanFrame & Frame);
      int  RingCount() const;
      bool RingBufferFull() const;
      const char * GetDev() const { return CanDriver ? CanDriver->dev : ""; }
      bool SendData(const KCanFrame & Frame);
      bool SendData(const KComfortFrame & Frame);
  
      virtual void Execute();
  };

#endif

