/*
 *
 *  Copyright (C) 2014 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__PYTHON__)
  #include <Python.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(__WINDOWS__)
  #include <windows.h>
#endif
#if defined(__LINUX__)
  #include <sys/socket.h>
  #include <netinet/in.h>
#endif

#include "KTcpCanClient.h"

const int SendTimeOut = 1000;
const int ReadTimeOut = 1000;

KTcpCanClient::KTcpCanClient()
{
  mConnected = false;
  Trace = false;
}

bool KTcpCanClient::SendBuffer(const char * aBuffer, int aLength, int TimeOut)
{
  if (!aBuffer || !aLength)
    return true;
  
  return SendBuf(aBuffer, aLength, 0, TimeOut) == aLength;
}

bool KTcpCanClient::WriteLine(const char * aBuffer)
{
  if (Trace && aBuffer)
  {
    const char * Ptr = aBuffer;
    do
    {
      const char * Next = strstr(Ptr, "\r\n");
      printf("Tcp send: ");
      if (Next)
      {
        for (int i = 0; i < Next-Ptr; i++)
          printf("%c", Ptr[i]);
        Ptr = Next + 2;
      } else {
        printf("%s", Ptr);
        Ptr = NULL;
      }
      printf("\n");
    } while(Ptr);
  }
  int len =  aBuffer ? (int) strlen(aBuffer) : 0;
  
  if (len)
    if (!SendBuffer(aBuffer, len, SendTimeOut))
      return false;
  
  return SendBuffer("\r\n", 2, len ? SendTimeOut : 100);
}

bool KTcpCanClient::ReadLine(char * aString, int aLength)
{
  const char * First = aString;
  aString[0] = 0;
   
  int Len = 0;
  char Temp[4];
 
  bool Ok = false;
  
  while (!Ok &&
         ReceiveBuf(Temp, 1, 0, Len ? 100 : ReadTimeOut) == 1)
  {
 /*   if (Temp[0] == 13 || Temp[0] == 10)
    {
      if (Trace && aLength > 2)
      {
        *aString++ = '\\';
        *aString++ = Temp[0] == 13 ? 'r' : 'n';
        *aString = 0;
        aLength += 2;
        Len += 2;
      }
    } */
    if (Temp[0] == 13)
    {
      
    } else
    if (Temp[0] == 10)
    {
      Ok = true;
    } else
    if (aLength <= 1)
      Ok = false;
    else {
      *aString = Temp[0];
      aString++;
      *aString = 0;
      aLength--;
      Len++;
    }
  }
 
  if (Ok && Trace)
    printf("Tcp recv: %s\n", First);
  return Ok;
}

void KTcpCanClient::Disconnect()
{
  if (mConnected)
  {
    // unter Windows SD_BOTH == 2
    // unter Linux SHUT_RDWR == 2
    if (mActive)
      ErrorCheck(shutdown(mSocket, 2), "shutdown");
    
    mConnected = false;
  }
  
  Close();
}

bool KTcpCanClient::Connect(int aTimeout)
{
  sockaddr_in addr;
  
  KIpSocket::Open();
  
  if (mActive && !mConnected)
  {
    GetSocketAddr(addr, mRemoteHost, mRemotePort);
    mConnected = ErrorCheck(connect(mSocket, (sockaddr *)(&addr), sizeof(addr)),
                            "connect") == 0;
    
    if (!mConnected && mBlockMode == bmNonBlocking && aTimeout > 0)
      Select(NULL, &mConnected, NULL, aTimeout);
    
    if (!mConnected)
    {
      Close();
    }
  }
  
  return mConnected;
}

bool KTcpCanClient::ReceiveBuffer(char * Buffer, int BufferLen)
{
  bool Ok = true;
  
  while (Ok && BufferLen > 0)
  {
    int getl = ReceiveBuf(Buffer, BufferLen, 0, 2000);
    Ok = getl > 0;
    if (Ok && Trace)
      for (int i = 0; i < getl; i++)
      {
        char c = Buffer[i];
        if (c != '\n' && (c < ' ' || c >= 127))
          c = '.';
        printf("%c", c);
      }
    Buffer += getl;
    BufferLen -= getl;
  }
  
  return Ok;  
}

bool KTcpCanClient::TraceReceive(int Length)
{
  bool Ok = true;
  char Buffer[1024];
  
  printf("Tcp recv: ");
  while (Ok && Length > 0)
  {
    int l = Length > (int) High(Buffer) ? High(Buffer) : Length;
    Ok = ReceiveBuffer(Buffer, l);
    Length -= l;
  }
  printf("\n");
  return Ok;
}

// Anfrage vom Internet Explorer:
//
// GET /path HTTP/1.1              << Unterverzeichnis hier einfügen
// Accept: */*
// Accept-Language: de-ch
// Accept-Encoding: gzip, deflate
// User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; ...
// Host: localhost:8080
// Connection: Keep-Alive
// leere Zeile

bool KTcpCanClient::ReadHttpStream(char * & aStream, int & aStreamLen, const char * aGetString)
{
  bool Ok = false;
  bool HeadOk = true;
  bool Chunked = false;
  char buffer[1024];
  char * stream = buffer;
  int stream_len = High(buffer);
  bool genResult = &aStream != NULL;
  bool connected = mConnected;
  
  if (&aStream != NULL)
    aStream = NULL;
  
  ContentLength = 0;
  if (connected || Connect(1000))
  {
    char Get[512];
    char remote[128];
    const char * ParsePtr;
    
    strcpy(Get, "GET ");
    strcat(Get, mRemotePath);
    strcat(Get, " HTTP/1.1\r\n");
    sprintf(remote, "Host: %s", mRemoteHost);
    Ok = WriteLine(aGetString ? aGetString : Get);
    if (Ok)
      Ok = WriteLine(remote);
    if (Ok)
      Ok = WriteLine("");
    if (Ok)
    {
      Ok = ReadLine(stream, stream_len);
      if (strncmp("HTTP/1.1 200 OK", stream, 14))
        HeadOk = false;
    }
    if (Ok)
    {
      // lesen bis zur ersten Leerzeile
      // (Header wird mit einer Leerzeile abgeschlossen)
      do
      {
        Ok = ReadLine(stream, stream_len);
        // "Content-Length: %d"
        if (Ok)
        {
          ParsePtr = strstr(stream, "Content-Length");
          if (ParsePtr)
          {
            ParsePtr += strlen("Content-Length");
            while (*ParsePtr && *ParsePtr <= ' ')
              ParsePtr++;
            if (*ParsePtr == ':')
            {
              ParsePtr++;
              ContentLength = (int)strtol(ParsePtr, NULL, 10);
            }
          } else
          if (strstr(stream, "Transfer-Encoding: chunked"))
            Chunked = true;
          else
          if (strstr(stream, "Connection: close"))
            connected = false;
        }
      } while (Ok && strlen(stream) != 0);
      
      if (0 < ContentLength)
      {
        if (genResult)
        {
          stream = aStream = new char[ContentLength + 1];
          stream_len = aStreamLen = ContentLength;
        } else
        if (ContentLength < stream_len)
          stream_len = ContentLength;
       
        if (!genResult && Trace)
          Ok = TraceReceive(stream_len);
        else 
          Ok = ReceiveBuffer(stream, stream_len);
      } else
      if (Chunked)
      {
        bool Ends = false;
        while (Ok && !Ends)
        {
          Ok = ReadLine(buffer, High(buffer)); // hex. Länge
            
          const char * Ptr = buffer;
          int Len = NCanUtils::GetHex(Ptr);
          Ends = !Len || *Ptr;
            
          if (Ok)
          {
            if (Trace)
              Ok = TraceReceive(Len + 2);
            else 
            if (genResult)
            {
              if (!aStream)
              {
                aStream = new char[Len + 2];
                aStreamLen = 0;
              } else {
                char * newStr = new char[aStreamLen + Len + 2];
                memcpy(newStr, aStream, aStreamLen);
                delete [] aStream;
                aStream = newStr;
              }
              Ok = ReceiveBuffer(aStream + aStreamLen, Len + 2);
              aStreamLen += Len;
            } else {
              Len += 2;
              while (Ok && Len > 0)
              {
                int l = Len > (int) High(buffer) ? (int) High(buffer) : Len;
                Ok = ReceiveBuffer(buffer, l);
                Len -= l;
              }
            }
          }
        } 
      }
    }
    
    if (!connected || !Ok)
      Disconnect();
    
    if (Trace)
      printf("\n");
    
    if (genResult && aStream)
    {
      if (Ok)
        aStream[aStreamLen] = 0;
      else {
        delete [] aStream;
        aStream = NULL;
        aStreamLen = 0;
      }
    }
  }
  
  return Ok && HeadOk;
}

bool KTcpCanClient::PutHttpStream(const char * aStream)
{
  bool Ok = false;
  bool HeadOk = true;
  char buffer[1024];
  char * stream = buffer;
  int stream_len = High(buffer);
  
  if (Connect(1000))
  {
    char Put[512];
    const char * ParsePtr;
    bool Chunked = false;
     
    sprintf(Put, "PUT /put_can_log.php HTTP/1.1\r\n"
                 "Host: %s\r\n"
                 "Content-Length: %d\r\n",
            mRemoteHost, (int) strlen(aStream) + 2);
    Ok = WriteLine(Put);
    if (Ok)
      Ok = WriteLine(aStream);
    
    if (Ok)
    {
      Ok = ReadLine(stream, stream_len);
      if (strncmp("HTTP/1.1 200 OK", stream, 14))
        HeadOk = false;
    }
    if (Ok)
    {
      // lesen bis zur ersten Leerzeile
      // (Header wird mit einer Leerzeile abgeschlossen)
      do
      {
        Ok = ReadLine(stream, stream_len);
        // "Content-Length: %d"
        if (Ok)
        {
          ParsePtr = strstr(stream, "Content-Length");
          if (ParsePtr)
          {
            ParsePtr += strlen("Content-Length");
            while (*ParsePtr && *ParsePtr <= ' ')
              ParsePtr++;
            if (*ParsePtr == ':')
            {
              ParsePtr++;
              ContentLength = (int)strtol(ParsePtr, NULL, 10);
            }
          } else
          if (strstr(stream, "Transfer-Encoding: chunked"))
            Chunked = true;
        }
      } while (Ok && strlen(stream) != 0);
      
      if (0 < ContentLength)
      {
        if (ContentLength < stream_len)
          stream_len = ContentLength;
       
        if (Trace)
          Ok = TraceReceive(stream_len);
      } else
      if (Chunked)
      {
        bool Ends = false;
        while (Ok && !Ends)
        {
          Ok = ReadLine(buffer, High(buffer)); // hex. Länge
            
          const char * Ptr = buffer;
          int Len = NCanUtils::GetHex(Ptr);
          Ends = !Len || *Ptr;
            
          if (Ok)
          {
            if (Trace)
              Ok = TraceReceive(Len + 2);
            else {
              if (Len > 0 && !Trace)
                printf("result: ");
              Len += 2;
              while (Ok && Len > 0)
              {
                int l = Len > (int) High(buffer) ? (int) High(buffer) : Len;
                Ok = ReceiveBuffer(buffer, l);                
                if (!Trace && Ok && Len > 2)
                {
                  buffer[l] = 0;
                  printf("%s", buffer);
                }
                Len -= l;
              }
            }
          }
        } 
      }
    }    
    Disconnect();
  }
  
  return Ok && HeadOk;
}

bool KTcpCanClient::ReadElster(char * Buffer)
{
  int Read = ReceiveBuf(Buffer, 1, 0, 500); // timeout: 500 ms
  if (Read == 1 && *Buffer != '\n')  // auf LF synchronisieren
  {
    for (int Tot = Read ; Tot < 3*12; Tot++)
    {
      Read = ReceiveBuf(Buffer + Tot, 1, 0, 5);
      if (Read != 1)
      {
        Buffer[Tot] = 0;
        printf("broken %s\n", Buffer);
        return false;
      }
      if (Buffer[Tot] == '\n')
      {
        Buffer[Tot + 1] = 0;
        return true;
      }
    }
    Buffer[3*12] = 0;
    printf("no LF %s\n", Buffer);
  }
  return false;
}

bool KTcpCanClient::SendElster(char * Buffer, int timeout)
{
  while (strlen(Buffer) && Buffer[strlen(Buffer)-1] == ' ')
    Buffer[strlen(Buffer)-1] = 0;

  strcat(Buffer, "\n");
  
  return SendBuffer(Buffer, (int) strlen(Buffer), timeout);
}

bool KTcpCanClient::SendComfortFrame(const KComfortFrame & send, KComfortFrame & recv)
{
  bool Ok = false;
  char str[64];

  send.SetToString(str);
  strcat(str, "\n");

  Connect(10);
  if (SendElster(str, 1000)) // 1000 ms timeout
    Ok = ReadElster(str);

  if (Ok)
    Ok = recv.SetFromString(str, true);

  Disconnect();

  return Ok;
}

