/*
 *
 *  Copyright (C) 2014 Jrg Mller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <vcl.h>
#pragma hdrstop

#if !defined(VCL)
  #define USE_KCOMM
#endif

#include <stdio.h>

#include "UCS_Dlg.h"
#if !defined(USE_KCOMM)
  #include "Comm.hpp"
#else
  #include "KComm.h"
  #define TComm KComm
#endif
#include "KTcpCanClient.h"
#include "NCanUtils.h"
#include "KElsterTable.h"
#include "KCanElster.h"
#include "KScanTable.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

bool Stop = false;
KCanElster * CanElster = NULL;
TComm * CS_Com = NULL;
TComm * Null_Com = NULL;
KTcpCanClient CanClient;

enum {dev_nullmodem, dev_can_server, dev_can232, dev_simulation } dev;

unsigned char not_connected_dev[256];

#if __BORLANDC__ >= 0x640
char * copy_tstr(const wchar_t * from)
#else
char * copy_tstr(const char * from)
#endif
{
  static char copy_str_char[1024];

  int i = 0;
  while (from[i])
  {
    copy_str_char[i] = from[i];
    i++;
  }
  copy_str_char[i] = 0;
  return copy_str_char;
}

bool ElsterComm(const KComfortFrame & send, KComfortFrame & recv);
unsigned short GetKennwort();

__fastcall TForm1::TForm1(TComponent* Owner)
  : TForm(Owner)                                                       
{
}

void __fastcall TForm1::FormCreate(TObject *Sender)
{
  char file_name[1024];
  
  cbDevice->ItemIndex = 3;
  dev = dev_can_server;
  pcMain->ActivePage = tsStart;
  edLog->Text = ExtractFilePath(Application->ExeName) + "cs_log.txt";
  
  int h = cbDevice->Top - cbComm->Top;
  lbRaspi->Top = lbDevice->Top + h;
  lbSecondCom->Top = lbDevice->Top + h;
  edRaspi->Top = cbDevice->Top + h;
  cbSecondCom->Top = cbDevice->Top + h;

  cbDeviceChange(NULL);

#if !defined(USE_KCOMM)
  CS_Com = new TComm(this);
#else
  CS_Com = new KComm;
#endif
  CS_Com->NewScreenSize(80, 2000);
  CS_Com->SetDTR(true);
  CS_Com->SetRTS(true);
  CS_Com->SetParity(2);
  CS_Com->SetStopBits(2);
  CS_Com->SetBaudrate(2400);
  CS_Com->SetRtsFlowControl(false);
  CS_Com->SetTimerInterval(100);
}

void __fastcall TForm1::FormShow(TObject *Sender)
{                      
  btnStart->SetFocus();
}

void TForm1::EnableControls(bool Ok)
{
  cbComm->Enabled = Ok;
  cbDevice->Enabled = Ok;
  edRaspi->Enabled = Ok;
  cbSecondCom->Enabled = Ok;
  cbxLog->Enabled = Ok;
  btnFileOpen->Enabled = Ok;
  edLog->Enabled = Ok;
}

void __fastcall TForm1::cbDeviceChange(TObject *Sender)
{
  if (0 <= cbDevice->ItemIndex && cbDevice->ItemIndex < 4)
  {
    dev = cbDevice->ItemIndex;
    bool UseCom = dev == dev_nullmodem ||
                  dev == dev_can232;
    lbRaspi->Visible = !UseCom && dev != dev_simulation;
    edRaspi->Visible = !UseCom && dev != dev_simulation;
    lbSecondCom->Visible = UseCom;
    cbSecondCom->Visible = UseCom;
    if (UseCom)
    {
      if (dev == dev_nullmodem)
        lbSecondCom->Caption = "2. Schnittstelle";
      else
        lbSecondCom->Caption = "can232";
    }
  }
}

bool ConnectedIndex(const KComfortFrame & cs, unsigned char & Index)
{
  KCanFrame Frame;

  bool Ok = cs.SetCanFrame(Frame);
  if (Ok)
    Ok = (Frame.Data[0] & 0x0f) == 1 &&
         Frame.Index() == 0x000b;

  Index = ShortCanId(Frame.ReceiverId());
 
  return Ok;
}

bool NotConnected(const KComfortFrame & cs)
{
  unsigned char Index;
  bool Ok = ConnectedIndex(cs, Index);
  
  if (Ok &&
      not_connected_dev[Index] > 1)
    return true;

  return false;
}

void __fastcall TForm1::btnStartClick(TObject *Sender)
{  
  if (btnStart->Caption == "Stopp")
  {
    Stop = true;
    return;
  }
  
  if (!InitCom())
  {
    CloseCom();
    return;
  }

  unsigned short Kennwort = 0;
  if (dev  == dev_can_server || dev == dev_can232 || dev == dev_simulation)
  {
    Kennwort = GetKennwort();
    if (Kennwort)
    {
      Kennwort = (Kennwort << 8) + (Kennwort >> 8);
      printf("Kennwort: %d\n", Kennwort);
    } else {
      printf("ungueltiges Kennwort\n");
      CloseCom();
      return;
    }
  }
  Stop = false;
  CS_Com->SetComPortNr(cbComm->ItemIndex + 1);
  try
  {
    CS_Com->InitCOM();
  }
  catch (Exception &exception)
  {
    Application->ShowException(&exception);
  }
  if (!CS_Com->COM_INITIALISIERT())
  {
    CS_Com->Show();
    CS_Com->write(" serielle Schnittstelle konnte nicht geoeffnet werden!  ");
    CloseCom();
    return;
  }

  btnStart->Caption = "Stopp";
  EnableControls(false);

  while (CS_Com->ReadFromCOM() != NULL_CHAR)   // Puffer leeren
    ;

  if (dev == dev_nullmodem)
    while (Null_Com->ReadFromCOM() != NULL_CHAR)
      ;

  KComfortFrame cs;
  KComfortFrame cs_recv;
  bool written = false;
  int len_cs = 0;
  int len_cs_recv = 0;
  do
  {
    CS_Com->CommMonitor(NULL);
    if (dev == dev_nullmodem)
    {
      Null_Com->CommMonitor(NULL);
      int res2 = Null_Com->ReadFromCOM();
      if (res2 < 0x100)
      {
        CS_Com->WriteToCOM((char)(res2));
        cs_recv.Data[len_cs_recv++] = (unsigned char)res2;
      } else
      if (res2 > 0x100)
        len_cs_recv = 0;

      if (len_cs_recv == 12)
      {
        KCanFrame Frame;
        char str[64];

        if (cbxUseElsterTable->Checked &&
            cs_recv.SetCanFrame(Frame) &&
            FormElsterTable(Frame, str))
          printf("Elster %3.3x.%4.4x.%4.4x: %s\n",
                 Frame.Id, Frame.Index(), Frame.Value(),str);
        if (cbxRecvTelegram->Checked)
        {
          cs_recv.SetToString(str);
          printf("recv: %s\n", str);
        }
        len_cs_recv = 0;
      }
    }
    int res = CS_Com->ReadFromCOM();
    if (res < 0x100)
    {
      if (dev == dev_nullmodem)
        Null_Com->WriteToCOM((char)(res));

      cs.Data[len_cs++] = char(res);
      if (cs.Data[0] != 0x0d)
        len_cs = 0;

      if (len_cs == 12)
      {
        KCanFrame SendFrame;
        char str[64];
        KComfortFrame::msg_type cf_type;

        if (dev == dev_nullmodem)
          len_cs_recv = 0;

        // cs ==> SendFrame und cf_type
        if (!cs.SetCanFrame(SendFrame))
          cf_type = KComfortFrame::mt_invalid;
        else
          cf_type = cs.GetMsgType();
        if (cf_type == KComfortFrame::mt_short_change)
          cf_type = KComfortFrame::mt_change;

        // show send frame 
        if (cf_type != KComfortFrame::mt_invalid)
        {
          if (cbxCanTelegram->Checked ||
              (cf_type == KComfortFrame::mt_change && !cbNotChange->Checked))
          {
            NCanUtils::FormatFrame(SendFrame, false, str);
            printf("send: %s", str);
          } else
          if (cbxCS_Telegram->Checked)
          {
            cs.SetToString(str);
            printf("send: %s\n", str);
          }
        }
        
        if ((cf_type == KComfortFrame::mt_change && cbNotChange->Checked))
        {
          if (FormElsterTable(SendFrame, str))
            printf("Elster %3.3x.%4.4x.%4.4x: %s\n",
                   SendFrame.Id, SendFrame.Index(), SendFrame.Value(),str);
        } else
        if (cf_type == KComfortFrame::mt_invalid)
        {
          cs.SetToString(str);
          printf("invalid: %s\n", str);
          cs_recv.SetOk(false);
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (dev == dev_nullmodem)
        {
          // do nothing
        } else
        if (cf_type == KComfortFrame::mt_init)
        {
          cs_recv.SetOkValue(Kennwort);
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (cf_type == KComfortFrame::mt_get &&
            NotConnected(cs))
        {
          cs_recv.SetOk(false);
          CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } else
        if (ElsterComm(cs, cs_recv)) // send and receive / interpret
        {
          KCanFrame Frame;
          bool show_recv = false;
          bool recv_ok = cs_recv.SetCanFrame(Frame);
          if (cbxUseElsterTable->Checked ||
              (cf_type == KComfortFrame::mt_change && !cbNotChange->Checked))
          {
            recv_ok = FormElsterTable(SendFrame, str);
            if (recv_ok)
              printf("Elster %3.3x.%4.4x.%4.4x: %s\n",
                     SendFrame.Id, SendFrame.Index(), SendFrame.Value(), str);
            else
            if (cs_recv.Data[0] != 0xaa && cs_recv.Data[0] != 0x55 &&
                !cbxRecvTelegram->Checked && cf_type != KComfortFrame::mt_init)
              show_recv = true;
          }
          if (cbxRecvTelegram->Checked || cf_type == KComfortFrame::mt_init || show_recv)
          {
            if (recv_ok &&cf_type != KComfortFrame::mt_init)
              NCanUtils::FormatFrame(Frame, false, str);
            else {
              cs_recv.SetToString(str);
              strcat(str, "\n");
            }
            if (cf_type == KComfortFrame::mt_init)
              printf("init: %s", str);
            else
              printf("recv: %s", str);
          }
          unsigned char con_ind;
          if (ConnectedIndex(cs, con_ind))
          {
            if (cs_recv.Data[0] == 0xaa)
              not_connected_dev[con_ind]++;
          }
          if (dev != dev_nullmodem)
            CS_Com->WriteBufferToCOM(cs_recv.Data, 12);
        } // ElsterComm
        len_cs = 0;
        written = true;
      }
      if (written)
      {
        Application->ProcessMessages();
        if (dev != dev_nullmodem)
          CS_Com->writeln();
        written = false;
      }
    } else
    if (res != NULL_CHAR)
    {
      len_cs = 0;
      CS_Com->WriteChar(13);
      CS_Com->WriteChar(10);
    } else
      NCanUtils::SleepMs(1);
    Application->ProcessMessages();
  } while (CS_Com->COM_INITIALISIERT() && !Stop);

  CloseCom();
  btnStart->Caption = "Start";
  EnableControls(true);
}

bool ElsterComm(const KComfortFrame & send, KComfortFrame & recv)
{
  bool Ok = false;
  try
  {
    switch (dev)
    {
      case dev_nullmodem:
        return true;

      case dev_can_server:
        Ok = CanClient.SendComfortFrame(send, recv);
        break;

      case dev_can232:
        if (!send.SetCanFrame(CanElster->SendFrame) ||
            !CanElster->Send(1, true))
          return false;
        return recv.GetCanFrame(CanElster->RecvFrame);

      case dev_simulation:
        Ok = Get_scan_value(send, recv);
        if (Ok && send.GetMsgType() == KComfortFrame::mt_change)
          recv.SetOk(true);
        NCanUtils::SleepMs(Ok ? 50 : 100);
        break;

      default:
        break;
    }
  }
  catch (Exception &exception)
  {
    Application->ShowException(&exception);
    Ok = false;
    if (dev == dev_can_server)
      CanClient.Disconnect();
    CS_Com->ExitCOM();
  }
  return Ok;
}

int GetElsterValue(unsigned short receiver_id, unsigned short elster_index)
{
  KComfortFrame send;
  KComfortFrame recv;
  KCanFrame Frame;

  Frame.InitElsterFrame(0x680, receiver_id, elster_index);
  Frame.Len = 7;
  send.GetCanFrame(Frame);
  bool Ok = ElsterComm(send, recv);
  if (Ok)
    Ok = recv.SetCanFrame(Frame);

  if (Ok)
    return Frame.Value();
  else
    return -1;
}

unsigned short GetKennwort()
{
  int res = GetElsterValue(0x601, 0x000b);
  if (res > 0)
    return (unsigned short) res;

  res = GetElsterValue(0x301, 0x000b);
  if (res > 0)
    return (unsigned short) res;

  // http://www.ip-symcon.de/forum/threads/15888-Wolf-CSZ-mit-ISM1
  // eventuell mit direkter RS232-Schnittststelle an WP
  res = GetElsterValue(0x0000, 0x015b); // GERAETEKONFIGURATION_2
  if (res > 0)
    return (unsigned short) res;

  return 0;
}

void TForm1::CloseCom()
{
  CS_Com->ExitCOM();
  if (CanElster)
  {
    CanElster->Halt();
    delete CanElster;
    CanElster = NULL;
  }
  if (Null_Com)
    Null_Com->ExitCOM();
}

bool TForm1::InitCom()
{
  bool Ok = true;

  dev = cbDevice->ItemIndex;

  if (dev == dev_can232)
  {
    char port[64];
    CanElster = new KCanElster;

    sprintf(port, "COM%d", cbSecondCom->ItemIndex + 1);
    if (!CanElster->Init(port))
      return false;

    CanElster->Trace = true;
    CanElster->Resume();

    return true;
  }

  if (dev == dev_can_server)
  {
    CanClient.SetRemoteHost(copy_tstr(edRaspi->Text.c_str()));
    CanClient.SetRemotePort(5524);

    return true;
  }

  if (dev == dev_nullmodem)
  {
    if (!Null_Com)
    {
#if !defined(USE_KCOMM)
      Null_Com = new TComm(this);
#else
      Null_Com = new KComm;
#endif
      Null_Com->NewScreenSize(80, 2000);
      Null_Com->SetDTR(true);
      Null_Com->SetRTS(true);
      Null_Com->SetParity(2);
      Null_Com->SetStopBits(2);
      Null_Com->SetBaudrate(2400);
      Null_Com->SetRtsFlowControl(false);
      Null_Com->SetTimerInterval(100);
    }
    try
    {
      Null_Com->SetComPortNr(cbSecondCom->ItemIndex + 1);
      Ok = Null_Com->InitCOM();
    }
    catch (Exception &exception)
    {
      Application->ShowException(&exception);
      Ok = false;
    }
  }

  if (dev == dev_simulation)
    return true;
    
  return Ok;
}

void __fastcall TForm1::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  CanClose = cbComm->Enabled;
}

void __fastcall TForm1::btnFileOpenClick(TObject *Sender)
{
  OpenDialog->FileName = edLog->Text;
  if (OpenDialog->Execute())
  {
    edLog->Text = OpenDialog->FileName;
  }
} 

void __fastcall TForm1::btnLoadScanClick(TObject *Sender)
{
  OpenDialog->FileName = edScanFile->Text;
  if (OpenDialog->Execute())
  {
    edScanFile->Text = OpenDialog->FileName;
  }
}

void __fastcall TForm1::btnLoadClick(TObject *Sender)
{
  if (LoadScanTable(copy_tstr(edScanFile->Text.c_str())))
    printf("Tabelle geladen\n");
  else
    printf("Tabelle nicht geladen\n");
}

